package com.bytesw.bytebot.service.agency;

import com.bytesw.bytebot.bean.AgentBean;
import com.bytesw.bytebot.bean.agency.*;
import com.bytesw.bytebot.model.Agent;
import com.bytesw.bytebot.model.agency.*;
import com.bytesw.bytebot.model.enums.AgentStatusEnum;
import com.bytesw.bytebot.repository.agency.*;
import com.bytesw.bytebot.service.CustomPaginationService;
import com.bytesw.xdf.dao.rsql.CustomRsqlVisitor;
import com.bytesw.xdf.exception.NotFoundException;
import com.bytesw.xdf.service.XDFService;
import com.bytesw.xdf.sql.beans.Pagination;
import com.bytesw.xdf.sql.beans.SortField;
import cz.jirutka.rsql.parser.RSQLParser;
import cz.jirutka.rsql.parser.ast.Node;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.stereotype.Service;

import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

@Service
public class FlightService extends XDFService<Flight, FlightBean, Integer>{

    @Autowired
    private AirlineRepository airlineRepository;

    @Autowired
    private AircraftRepository aircraftRepository;

    @Autowired
    private AirportRepository airportRepository;

    @Autowired
    private FlightRepository flightRepository;


    protected FlightService(FlightRepository repository) {
        super(repository);
    }

    @Override
    protected Flight toModel(Flight flight, FlightBean flightBean) {
        return super.toModel(flight, flightBean);
    }

    @Override
    protected FlightBean toBean(Flight model) {
        return FlightBean.miniClone(model);
    }

    @Override
    public FlightBean getById(Integer id) {
        Optional flight = repository.findById(id);

        if (!flight.isPresent())
            throw new NotFoundException("Flight not found!");

        return FlightBean.miniClone((Flight) flight.get());
    }

    public Sort createSort(Pagination pagination) {
        List<Sort.Order> orderList = new ArrayList<>();
        for (SortField sortField : pagination.getSortFields()) {
            if ("asc".equalsIgnoreCase(sortField.getDirection())) {
                orderList.add(Sort.Order.asc(sortField.getField()));
            } else {
                orderList.add(Sort.Order.desc(sortField.getField()));
            }
        }
        return Sort.by(orderList);
    }

    public Specification<Flight> createSpecification(final String filterExpression) {
        Node rootNode = new RSQLParser().parse(filterExpression);
        return rootNode.accept(new CustomRsqlVisitor<>());
    }

    @Override
    public void searchByPagination(Pagination<FlightBean> pagination) {
        if (pagination.getItemsPerPage() == 0) {
            pagination.setItemsPerPage(10);
        }

        Sort sort = null;
        if (pagination.getSortFields() != null) {
            sort = createSort(pagination);
        }

        Pageable pageable = null;
        if (sort != null) {
            pageable = PageRequest.of(pagination.getCurrentPage(), pagination.getItemsPerPage(), sort);
        } else {
            pageable = PageRequest.of(pagination.getCurrentPage(), pagination.getItemsPerPage());
        }

        Specification<Flight> specification = (Specification<Flight>) (root, query, criteriaBuilder)
                -> criteriaBuilder.isNull(root.get("deletedAt"));

        if (pagination.getFilterExpression() != null && pagination.getFilterExpression().length() > 0) {
            specification = createSpecification(pagination.getFilterExpression()).and(specification);
        }

        Page<Flight> page = flightRepository.findAll(specification, pageable);

        pagination.setTotalItems(page.getTotalElements());
        pagination.setTotalPages(page.getTotalPages());

        // Clone to array
        List<FlightBean> list = new ArrayList<>();

        for (Flight flight : page.getContent()) {
            list.add(FlightBean.miniClone(flight));
        }

        pagination.setData(list);
    }

    public List<FlightSearchBean> searchFlights(Integer airportOriginId, Integer airportDestinationId, String departureDate) {
        System.out.println(departureDate);
        LocalDateTime startDate = LocalDateTime.parse(departureDate);
        List<Flight> flights = flightRepository.searchFlights(airportOriginId, airportDestinationId, startDate);

        return flights.stream()
                .map(flight -> FlightSearchBean.miniClone(flight, startDate)).collect(Collectors.toList());
    }

    @Override
    public FlightBean create(FlightBean flightBean) {

        DateTimeFormatter formatter = DateTimeFormatter.ofPattern("yyyy-MM-dd'T'HH:mm:ss");
        Optional<Airline> airline = airlineRepository.findById(flightBean.getAirlineId());
        Optional<Aircraft> aircraft = aircraftRepository.findById(flightBean.getAircraftId());

        if (!airline.isPresent())
            throw new NotFoundException("Airline not found!");

        if (!aircraft.isPresent())
            throw new NotFoundException("Aircraft not found!");

        LocalDateTime departureDate = LocalDateTime.parse(flightBean.getDepartureDate());
        LocalDateTime arrivalDate = LocalDateTime.parse(flightBean.getArrivalDate());

        if (departureDate.isAfter(arrivalDate) || departureDate.equals(arrivalDate))
            throw new IllegalArgumentException("Departure date invalid!");

        if (flightBean.getScales().isEmpty())
            throw new IllegalArgumentException("Scales can't be empty!");

        if (flightBean.getCosts().isEmpty())
            throw new IllegalArgumentException("Costs can't be empty!");

        Flight flight = new Flight();
        List<ScaleBean> scales = flightBean.getScales();
        List<Scale> scalesModel = new ArrayList<>();

        for (ScaleBean scale: scales) {
            Optional<Airport> airportOrigin = airportRepository.findById(scale.getAirportOriginId());
            Optional<Airport> airportDestination = airportRepository.findById(scale.getAirportDestinationId());

            if (!airportOrigin.isPresent() || !airportDestination.isPresent())
                throw new NotFoundException("Airport not found!");

            LocalDateTime scaleDeparture = LocalDateTime.parse(scale.getDepartureDate());
            LocalDateTime scaleArrival = LocalDateTime.parse(scale.getArrivalDate());

            if (scaleDeparture.isAfter(scaleArrival) ||
                    scaleDeparture.isEqual(scaleArrival))
                throw new IllegalArgumentException("Departure date invalid!");

            Scale scaleModel = new Scale();
            scaleModel.setAirportOrigin(airportOrigin.get());
            scaleModel.setAirportDestination(airportDestination.get());
            scaleModel.setDepartureDate(scaleDeparture);
            scaleModel.setArrivalDate(scaleArrival);
            scaleModel.setFlight(flight);

            scalesModel.add(scaleModel);
        }

        List<CostBean> costs = flightBean.getCosts();
        List<Cost> costModel = new ArrayList<>();

        for (CostBean cost: costs) {
            LocalDateTime costDeparture = LocalDateTime.parse(cost.getDepartureDate());
            LocalDateTime costArrival = LocalDateTime.parse(cost.getArrivalDate());
            if (costDeparture.isAfter(costArrival) ||
                costDeparture.isEqual(costArrival))
                throw new IllegalArgumentException("Departure date invalid!");

            if (cost.getAmount() == 0)
                throw new IllegalArgumentException("Cost can't be zero!");

            Cost oneCost = new Cost();
            oneCost.setAmount(cost.getAmount());
            oneCost.setDepartureDate(costDeparture);
            oneCost.setArrivalDate(costArrival);
            oneCost.setFlight(flight);

            costModel.add(oneCost);
        }

        flight.setAirline(airline.get());
        flight.setAircraft(aircraft.get());
        flight.setDepartureDate(departureDate);
        flight.setArrivalDate(arrivalDate);
        flight.setScales(scalesModel);
        flight.setCosts(costModel);
        flight.setAirportOrigin(flight.getScales().get(0).getAirportOrigin());
        flight.setAirportDestination(flight.getScales().get(flight.getScales().size() - 1).getAirportDestination());

        return FlightBean.miniClone(flightRepository.save(flight));
    }

    public FlightBean update(FlightBean flightBean, Integer id) {
        Optional<Flight> flight = flightRepository.findById(id);

        if (!flight.isPresent())
            throw new NotFoundException("Flight not found!");

        Flight flightEdit = flight.get();

        LocalDateTime departureDate = LocalDateTime.parse(flightBean.getDepartureDate());
        LocalDateTime arrivalDate = LocalDateTime.parse(flightBean.getArrivalDate());

        List<Cost> costsNew = flightBean.getCosts().stream().map(costBean -> {
            Cost cost = new Cost();
            cost.setFlight(flightEdit);
            cost.setAmount(costBean.getAmount());
            cost.setDepartureDate(LocalDateTime.parse(costBean.getDepartureDate()));
            cost.setArrivalDate(LocalDateTime.parse(costBean.getArrivalDate()));
            return cost;
        }).collect(Collectors.toList());

        flightEdit.getCosts().addAll(costsNew);
        flightEdit.setDepartureDate(departureDate);
        flightEdit.setArrivalDate(arrivalDate);

        return FlightBean.miniClone(flightRepository.save(flightEdit));
    }

    public void delete(Integer id) {

        Optional<Flight> flight = flightRepository.findById(id);

        if (!flight.isPresent())
            throw new NotFoundException("Flight not found!");

        flight.get().setDeletedAt(LocalDate.now());

        flightRepository.save(flight.get());
    }
}
