package com.bytesw.bytebot.service.agency;

import com.bytesw.bytebot.bean.agency.FlightSearchBean;
import com.bytesw.bytebot.bean.agency.ReservationBean;
import com.bytesw.bytebot.bean.agency.SupervisorBean;
import com.bytesw.bytebot.bean.agency.UpdateReservationStatus;
import com.bytesw.bytebot.model.agency.*;
import com.bytesw.bytebot.model.agency.enums.PaymentStatusEnum;
import com.bytesw.bytebot.model.agency.enums.ReservationStatusEnum;
import com.bytesw.bytebot.model.agency.enums.TicketTypeEnum;
import com.bytesw.bytebot.model.agency.enums.TravelClassEnum;
import com.bytesw.bytebot.repository.agency.AgencyAgentRepository;
import com.bytesw.bytebot.repository.agency.FlightRepository;
import com.bytesw.bytebot.repository.agency.PassengerRepository;
import com.bytesw.bytebot.repository.agency.ReservationRepository;
import com.bytesw.bytebot.service.CustomPaginationService;
import com.bytesw.bytebot.utils.CustomPaginationBean;
import com.bytesw.xdf.exception.NotFoundException;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.Pageable;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.stereotype.Service;

import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

@Service
public class ReservationService {

    @Autowired
    private ReservationRepository reservationRepository;

    @Autowired
    private FlightRepository flightRepository;

    @Autowired
    private PassengerRepository passengerRepository;

    @Autowired
    private AgencyAgentRepository agentRepository;

    public CustomPaginationBean<ReservationBean> getByPaginationAgent(CustomPaginationBean<ReservationBean> pagination) {
        Pageable pageable = CustomPaginationBean.buildPageable(pagination);
        Specification<Reservation> specification = (Specification<Reservation>) (root, query, criteriaBuilder)
                -> criteriaBuilder.isNull(root.get("deletedAt"));

        Page<Reservation> page = reservationRepository.findAll(specification, pageable);

        Page<ReservationBean> supervisorBeans = convertPage(page);

        return CustomPaginationBean.fromPage(supervisorBeans);
    }

    public Page<ReservationBean> convertPage(Page<Reservation> originalPage) {
        List<ReservationBean> reservationBean = originalPage.getContent().stream()
                .map(ReservationBean::miniClone)
                .collect(Collectors.toList());

        return new PageImpl<>(reservationBean, originalPage.getPageable(), originalPage.getTotalElements());
    }

    public ReservationBean create(ReservationBean reservationBean) {
        Reservation reservation = new Reservation();

        Optional<Passenger> passengerOptional = passengerRepository.findByIdAndDeletedAtIsNull(reservationBean.getPassengerId());

        if (!passengerOptional.isPresent())
            throw new NotFoundException("Passenger not found!!");

        Optional<AgencyAgent> agentOptional = agentRepository.findByIdAndDeletedAtIsNull(reservationBean.getAgentId());

        if (!agentOptional.isPresent())
            throw new NotFoundException("Agent not found!!");

        Optional<Flight> flightOptional = flightRepository.findById(reservationBean.getFlightId());

        if (!flightOptional.isPresent())
            throw new NotFoundException("Flight not found!!");

        reservation.setPassenger(passengerOptional.get());
        reservation.setAgencyAgent(agentOptional.get());
        reservation.setStatus(ReservationStatusEnum.RESERVED);
        reservation.setDepartureDate(LocalDateTime.parse(reservationBean.getDepartureDate()));
        reservation.setTicketType(TicketTypeEnum.fromString(reservationBean.getTicketType()));
        reservation.setTravelClass(TravelClassEnum.fromString(reservationBean.getTravelClass()));
        reservation.setFlight(flightOptional.get());
        reservation.setCreatedDate(LocalDate.now());

        Payment payment = new Payment();

        payment.setStatus(PaymentStatusEnum.WAITING);
        payment.setDate(LocalDate.now());
        payment.setAmount(FlightSearchBean.getAmount(flightOptional.get().getCosts(), LocalDateTime.parse(reservationBean.getDepartureDate())));

        reservation.setPayment(payment);

        return ReservationBean.miniClone(reservationRepository.save(reservation));
    }

    public ReservationBean getById(Integer id) {
        Optional<Reservation> reservation = reservationRepository.findByIdAndDeletedAtIsNull(id);

        if (!reservation.isPresent())
            throw new NotFoundException("Reservation not found!");

        return ReservationBean.miniClone(reservation.get());
    }

    public ReservationBean updateStatusReservation(UpdateReservationStatus status, Integer id) {
        Optional<Reservation> reservation = reservationRepository.findByIdAndDeletedAtIsNull(id);

        if (!reservation.isPresent())
            throw new NotFoundException("Reservation not found!!");

        ReservationStatusEnum statusEnum = ReservationStatusEnum.fromString(status.getStatus());

        reservation.get().setStatus(statusEnum);

        if (statusEnum == ReservationStatusEnum.PAID) {
            reservation.get().getPayment().setStatus(PaymentStatusEnum.PAID);
            reservation.get().getPayment().setDate(LocalDate.now());
        } else if (statusEnum == ReservationStatusEnum.CANCELED) {
            reservation.get().getPayment().setStatus(PaymentStatusEnum.CANCELED);
            reservation.get().getPayment().setDate(LocalDate.now());
        } else {
            throw new NotFoundException("Status not found!");
        }

        return ReservationBean.miniClone(reservationRepository.save(reservation.get()));
    }

    public ReservationBean delete(Integer id) {
        Optional<Reservation> reservationOptional = reservationRepository.findByIdAndDeletedAtIsNull(id);

        if (!reservationOptional.isPresent())
            throw new NotFoundException("Reservation not found!");

        reservationOptional.get().setDeletedAt(LocalDate.now());

        return ReservationBean.miniClone(reservationRepository.save(reservationOptional.get()));
    }
}
